# Sentinel object indicating the default broker should be used.
# Allows passing `None` to deactivate broker connections in develop and test
# environments.
DEFAULT = object()

# The connection settings used for connecting to RabbitMQ broker.
# Default settings are for a local installation of RabbitMQ.
BROKER = 'amqp://guest:guest@localhost:5672/%2F'



from datetime import datetime
import json
from uuid import uuid4

FACTOR = 10**6


def to_timestamp(dt, epoch=datetime(1970, 1, 1)):
    """http://stackoverflow.com/questions/8777753/converting-datetime-date-to-utc-timestamp-in-python"""
    td = dt - epoch
    return (td.microseconds + (td.seconds + td.days * 86400) * FACTOR) / FACTOR


class DomainEvent(object):

    def __init__(self,
                 routing_key=u"",
                 data={},
                 domain_object_id=None,
                 uuid_string=None,  # only set if recreated from json repr
                 timestamp=None,  # only set if recreated from json repr
                 retries=0,
                 **kwargs):
        """
        Define a Domain Event.

        :param str routing_key: The routing key is of the form
            ``<DOMAIN>.<EVENT_TYPE>``.  The routing key should be a descriptive
            name of the domain event such as ``user.registered``.
        :param dict data: The actual event data. *Must* be json serializable.
        :param str domain_object_id: Domain identifier of the event. This field
            is optional. If used, it might make search in an event store easier.
        :param str uuid_string: This UUID identifier of the event. If left
            ``None``, a new one will be created.
        :param float timestamp: Unix timestamp. If timestamp is None, a new
            (UTC) timestamp will be created.
        :param int retries: Number of times this event was delivered to a
            subscriber already.
        """
        self.routing_key = routing_key
        self.data = data
        self.domain_object_id = domain_object_id
        self.uuid_string = uuid_string
        if uuid_string is None:
            self.uuid_string = str(uuid4())
        self.timestamp = timestamp
        if timestamp is None:
            timestamp = to_timestamp(datetime.utcnow())
            self.timestamp = timestamp
        event_data = self.__dict__.copy()
        self.event_data = event_data
        self.retries = retries

    @classmethod
    def from_json(cls, json_data):
        """
        Create a DomainEvent from ``json_data``. Note that you probably want to
        dispatch first based on domain and event type.

        :param str json_data: Serialized domain event data.
        :rtype: DomainEvent
        """
        try:
            json_data = json_data.decode('utf-8')
        except AttributeError:
            pass
        return cls(**json.loads(json_data))

    def __repr__(self):
        return u"DomainEvent('{0.routing_key}', '{0.data}', (domain_obj_id: {0.domain_object_id}, uuid: {0.uuid_string}))".format(self) # noqa

    __str__ = __repr__

    def __eq__(self, other):
        return self.event_data == other.event_data



from django.apps import AppConfig
from django.conf import settings as djsettings

from domain_event_broker import settings


class DomainEventsConfig(AppConfig):
    name = 'domain_event_broker.django'
    label = 'domain_event_broker_django'
    verbose_name = 'Domain Events'

    def ready(self):
        if hasattr(djsettings, 'DOMAIN_EVENT_BROKER'):
            settings.BROKER = djsettings.DOMAIN_EVENT_BROKER



from time import sleep
from domain_event_broker import Publisher, Subscriber, Retry, publish_domain_event
from .helpers import (
    check_queue_exists, get_message_from_queue, get_queue_size,
    )
import uuid

from .. import settings


def nop(event):
    pass


class ConsumerError(Exception):
    pass


def raise_error(event):
    raise ConsumerError("Unexpected error")


def test_retry():
    def raise_retry(event):
        raise_retry.received += 1
        raise Retry(0.1)
    raise_retry.received = 0

    name = 'test-retry'
    subscriber = Subscriber()
    subscriber.register(raise_retry, name, ['test.retry'], max_retries=3)
    data = dict(message=str(uuid.uuid4())[:4])
    publish_domain_event('test.retry', data)
    subscriber.start_consuming(timeout=1.0)
    assert raise_retry.received == 4
    assert get_queue_size(name) == 0


def test_retry_delays():
    # Events with shorter retry delays should be processed before the ones with
    # longer delays.
    def raise_retry(event):
        raise_retry.received.append(event.data['delay'])
        raise Retry(event.data['delay'])
    raise_retry.received = []

    name = 'test-retry-delay'
    subscriber = Subscriber()
    subscriber.register(raise_retry, name, ['test.retry-delay'], max_retries=1)
    publish_domain_event('test.retry-delay', {'delay': 0.3})
    publish_domain_event('test.retry-delay', {'delay': 0.1})
    subscriber.start_consuming(timeout=1.0)
    assert raise_retry.received == [0.3, 0.1, 0.1, 0.3]
    assert get_queue_size(name) == 0


def test_auto_delete():
    name = 'test-auto-delete'
    subscriber = Subscriber()
    subscriber.register(nop, name, ['test.auto_delete'], auto_delete=True)
    publish_domain_event('test.auto_delete', {})
    subscriber.start_consuming(timeout=1.0)
    assert not check_queue_exists(name)


def test_multiple_listeneres():
    def one(event):
        one.received += 1
    one.received = 0

    def two(event):
        two.received += 1
    two.received = 0

    subscriber = Subscriber()
    subscriber.register(one, 'test-listener-one', ['test.one'])
    subscriber.register(two, 'test-listener-two', ['test.two'])
    publish_domain_event('test.one', {})
    publish_domain_event('test.two', {})
    subscriber.start_consuming(timeout=1.0)
    assert one.received == 1
    assert two.received == 1


def test_consumer_timeout():
    name = 'test-consumer-timeout'
    subscriber = Subscriber()
    subscriber.register(nop, name, ['#'])
    subscriber.start_consuming(timeout=1.0)


def test_dead_letter():
    name = 'test-dead-letter'
    subscriber = Subscriber()
    subscriber.register(raise_error, name, ['test.dl'], dead_letter=True)
    data = dict(message=str(uuid.uuid4())[:4])
    publish_domain_event('test.dl', data)
    subscriber.start_consuming(timeout=1.0)
    header, event = get_message_from_queue('test-dead-letter-dl')
    assert event.data == data
    transport = Subscriber()
    transport.channel.queue_delete(queue='test-dead-letter')
    transport.channel.queue_delete(queue='test-dead-letter-dl')


def test_invalid_json():
    name = 'test-invalid-json'
    subscriber = Subscriber()
    subscriber.register(nop, name, ['#'])
    publisher = Publisher(settings.DEFAULT)
    publisher.publish('iamnotvalidjson[]', 'foo')
    publisher.disconnect()
    subscriber.start_consuming(timeout=1.0)


def test_long_running_consumer():
    # Test a job that takes longer than the heartbeat of the connection.
    # Make sure the consumer finishes processing and acknowledges the event.
    name = 'test-long-job'

    def slow_nop(event):
        sleep(3.0)
        slow_nop.finished += 1

    slow_nop.finished = 0
    subscriber = Subscriber(connection_settings=settings.BROKER+'?heartbeat=1')
    subscriber.register(slow_nop, name, ['test.long_job'])
    publish_domain_event('test.long_job', {})
    subscriber.start_consuming(timeout=5.0)
    assert get_queue_size(name) == 0
    assert slow_nop.finished == 1



from django.core.management import call_command
from ..helpers import get_message_from_queue, get_queue_size


def test_replay(dead_letter_message):
    call_command('replay_domain_event', 'test-replay')
    assert get_queue_size('test-replay-dl') == 0
    header, event = get_message_from_queue('test-replay')
    assert event.data == dead_letter_message



import json
from unittest.mock import patch

from domain_event_broker import DomainEvent, publish_domain_event


@patch('domain_event_broker.transport.Publisher.publish')
class TestEvent(object):

    def test_fire_event(self, mock):
        """Use basic API to fire event"""
        event = publish_domain_event('test.test', {})
        assert mock.called
        json_data, routing_key = mock.call_args[0]
        event_data = json.loads(json_data)
        assert routing_key == "test.test"
        assert event_data['data'] == {}
        assert event_data['uuid_string'] == event.uuid_string

    def test_event_loading(self, mock):
        event = publish_domain_event('test.test', {})
        assert mock.called
        json_data, routing_key = mock.call_args[0]
        new_event = DomainEvent.from_json(json_data)
        assert new_event == event
